"""Module for detecting pspy services from parsing Python code."""

import ast


def parse_file(py_file):
  visitor = NodeVisitor(py_file)
  f = open(py_file, 'r')
  node = ast.parse(f.read())
  visitor.visit(node)
  return visitor.get_services_by_provider()


def parse_string(string):
  visitor = NodeVisitor()
  node = ast.parse(string)
  visitor.visit(node)
  return visitor.get_services_by_provider()


class NodeVisitor(ast.NodeVisitor):

  NIL_KWARGS = ('', {}, [], '')

  def __init__(self, py_file=None):
    self.services_by_provider = {}
    self.py_prefix = (py_file.replace('/', '.')[:-3] + '.') if py_file else ''

  def get_services_by_provider(self):
    """Get the services organized by service provider.

    Returns a dictionary of service provider -> (services, groups, perms,
    additional_groups).
    """
    return self.services_by_provider

  def visit(self, node):
    method = 'visit_%s' % node.__class__.__name__
    return getattr(self, method, self.generic_visit)(node)

  def generic_visit(self, node):
    ast.NodeVisitor.generic_visit(self, node)

  def visit_ClassDef(self, node):
    self.__get_services(node)

  def visit_FunctionDef(self, node):
    self.__get_services(node)

  def __get_services(self, node):
    for service in node.decorator_list:
      if hasattr(service, 'func'):
        check = ''
        if hasattr(service.func, 'attr'):
          check = service.func.attr
        else:
          check = service.func.id
        if check == 'service':
          kwargs = self.__parse_args(service)
          fqnn = self.__fully_qualify_node_name(node.name)
          service_provider = kwargs[3]
          if service_provider:
            if service_provider in self.services_by_provider:
              l = self.services_by_provider[service_provider]
              l[0].append(fqnn)  # services
              l[2].update(kwargs[1])  # perms
              l[3].extend(kwargs[2])  # additional groups
            else:
              self.services_by_provider[service_provider] = (
                  [fqnn], kwargs[0], kwargs[1], kwargs[2])
          else:
            self.services_by_provider[fqnn] = (
                [fqnn], kwargs[0], kwargs[1], kwargs[2])

  def __parse_args(self, node):
    ret = self.NIL_KWARGS
    if hasattr(node, 'keywords'):
      kwargs = {}
      for k in node.keywords:
        kwargs[k.arg] = ast.literal_eval(k.value)
      ret = (kwargs['group'] if 'group' in kwargs else '',
             kwargs['perms'] if 'perms' in kwargs else {},
             kwargs['additional_groups'] if 'additional_groups' in kwargs else [],
             kwargs['service_provider'] if 'service_provider' in kwargs else '')
    return ret

  def __fully_qualify_node_name(self, node_name):
    return self.py_prefix + node_name
