import config
import os
import rpclib
import socket


def service(group='', perms={}, additional_groups=[], service_provider=''):
  """Decorates a pspy service.

  Args:
    group: optional string specifying the group this service should
      belong to. Defaults to '' which means that the service won't share the
      group with any other service.
    perms: optional dict of filename keys to permission values. The permission
      values should be in octal form. Defaults to the empty dict.
    additional_groups: optional list of group names that the service should
      run under. Defaults to empty list.
    service_provider: optional string specifying the provider for this service.
      Defaults to '' which means that the service will have its own unique
      provider.

  Usage:
     @service(group='DatabaseGroup', {'db.sql': 0770}, ['group1', 'group2'])
     class BankService(object):
        ...

     or

     @service()
     def transfer(amount, dst, sender):
         ...
  """
  def decorator(obj):
    name = service_provider
    if service_provider == "":
      name = obj.__name__
    if config.server == name:
      print '[INFO] I am the server side of service %s provided by %s' % (obj.__name__, name)
      return obj
    print '[INFO] I am the client stub for service %s' % obj.__name__
    if hasattr(obj, '__call__'):
      return __client_fn(obj, service_provider)
    else:
      return __client_cls(obj, service_provider)
  return decorator


def __client_fn(fn, service_provider):
  if service_provider == '':
    service_provider = fn.__name__
  def client_call(fn):
    def call(*args, **kwargs):
      if not hasattr(fn, 'client'):
        fn.client = ServiceProviderClient(service_provider)
      fn.client.send_req(fn.__name__, fn.__name__, *args, **kwargs)
      return fn.client.rcv_req()
    return call
  return client_call(fn)


def __client_cls(cls, service_provider):
  if service_provider == '':
    service_provider = cls.__name__
  def client_call(method):
    def call(self, *args, **kwargs):
      if not hasattr(self, 'client'):
        self.client = ServiceProviderClient(service_provider)
      self.client.send_req(cls.__name__, method, *args, **kwargs)
      return self.client.rcv_req()
    return call

  for m in cls.__dict__:
    setattr(cls, m, client_call(m))
  return cls


class ServiceProviderClient:

  def __init__(self, service_provider):
    self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sockpath = os.path.join('sock-' + service_provider, 'sock')
    self.sock.connect(sockpath)
    print '[INFO] Client connected to sock %s' % sockpath
    self.lines = rpclib.buffered_readlines(self.sock)

  def send_req(self, service, method, *args, **kwargs):
    self.sock.sendall(
        rpclib.format_req(service, method, *args, **kwargs) + '\n')

  def rcv_req(self):
    return rpclib.parse_resp(self.lines.next())
